@@ -4,12 +4,12 @@ import (
 	"context"
 	"fmt"
 	"net/url"
+	"path"
 	"reflect"
 	"strconv"
+	"strings"
 	"time"
 
-	sdkapi "kubevirt.io/controller-lifecycle-operator-sdk/pkg/sdk/api"
-
 	"github.com/go-logr/logr"
 	"github.com/pkg/errors"
 	corev1 "k8s.io/api/core/v1"
@@ -26,11 +26,12 @@ import (
 	"sigs.k8s.io/controller-runtime/pkg/reconcile"
 	"sigs.k8s.io/controller-runtime/pkg/source"
 
-	cdiv1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
+	cdiv1 "kubevirt.io/containerized-data-importer-api/pkg/apis/core/v1beta1"
 	"kubevirt.io/containerized-data-importer/pkg/common"
 	featuregates "kubevirt.io/containerized-data-importer/pkg/feature-gates"
 	"kubevirt.io/containerized-data-importer/pkg/util"
 	"kubevirt.io/containerized-data-importer/pkg/util/naming"
+	sdkapi "kubevirt.io/controller-lifecycle-operator-sdk/pkg/sdk/api"
 )
 
 const (
@@ -61,6 +62,10 @@ const (
 	AnnCertConfigMap = AnnAPIGroup + "/storage.import.certConfigMap"
 	// AnnContentType provides a const for the PVC content-type
 	AnnContentType = AnnAPIGroup + "/storage.contentType"
+	// AnnRegistryImportMethod provides a const for registry import method annotation
+	AnnRegistryImportMethod = AnnAPIGroup + "/storage.import.registryImportMethod"
+	// AnnRegistryImageStream provides a const for registry image stream annotation
+	AnnRegistryImageStream = AnnAPIGroup + "/storage.import.registryImageStream"
 	// AnnImportPod provides a const for our PVC importPodName annotation
 	AnnImportPod = AnnAPIGroup + "/storage.import.importPodName"
 	// AnnRequiresScratch provides a const for our PVC requires scratch annotation
@@ -75,12 +80,19 @@ const (
 	AnnThumbprint = AnnAPIGroup + "/storage.import.vddk.thumbprint"
 	// AnnPreallocationApplied provides a const for PVC preallocation annotation
 	AnnPreallocationApplied = AnnAPIGroup + "/storage.preallocation"
+	// AnnExtraHeaders provides a const for our PVC extraHeaders annotation
+	AnnExtraHeaders = AnnAPIGroup + "/storage.import.extraHeaders"
+	// AnnSecretExtraHeaders provides a const for our PVC secretExtraHeaders annotation
+	AnnSecretExtraHeaders = AnnAPIGroup + "/storage.import.secretExtraHeaders"
 
 	//LabelImportPvc is a pod label used to find the import pod that was created by the relevant PVC
 	LabelImportPvc = AnnAPIGroup + "/storage.import.importPvcName"
 	//AnnDefaultStorageClass is the annotation indicating that a storage class is the default one.
 	AnnDefaultStorageClass = "storageclass.kubernetes.io/is-default-class"
 
+	// AnnOpenShiftImageLookup is the annotation for OpenShift image stream lookup
+	AnnOpenShiftImageLookup = "alpha.image.policy.openshift.io/resolve-names"
+
 	// ErrImportFailedPVC provides a const to indicate an import to the PVC failed
 	ErrImportFailedPVC = "ErrImportFailed"
 	// ImportSucceededPVC provides a const to indicate an import to the PVC failed
@@ -91,6 +103,10 @@ const (
 
 	// ImportTargetInUse is reason for event created when an import pvc is in use
 	ImportTargetInUse = "ImportTargetInUse"
+
+	// importPodImageStreamFinalizer ensures image stream import pod is deleted when pvc is deleted,
+	// as in this case pod has no pvc OwnerReference
+	importPodImageStreamFinalizer = "cdi.kubevirt.io/importImageStream"
 )
 
 // ImportReconciler members
@@ -117,6 +133,8 @@ type importPodEnvVar struct {
 	certConfigMap      string
 	diskID             string
 	uuid               string
+	readyFile          string
+	doneFile           string
 	backingFile        string
 	thumbprint         string
 	filesystemOverhead string
@@ -129,6 +147,22 @@ type importPodEnvVar struct {
 	httpsProxy         string
 	noProxy            string
 	certConfigMapProxy string
+	extraHeaders       []string
+	secretExtraHeaders []string
+}
+
+type importerPodArgs struct {
+	image                   string
+	importImage             string
+	verbose                 string
+	pullPolicy              string
+	podEnvVar               *importPodEnvVar
+	pvc                     *corev1.PersistentVolumeClaim
+	scratchPvcName          *string
+	podResourceRequirements *corev1.ResourceRequirements
+	workloadNodePlacement   *sdkapi.NodePlacement
+	vddkImageName           *string
+	priorityClassName       string
 }
 
 // NewImportController creates a new instance of the import controller.
@@ -196,6 +230,10 @@ func isPVCComplete(pvc *corev1.PersistentVolumeClaim) bool {
 	return exists && (phase == string(corev1.PodSucceeded))
 }
 
+func isImageStream(pvc *corev1.PersistentVolumeClaim) bool {
+	return pvc.Annotations[AnnRegistryImageStream] == "true"
+}
+
 // Reconcile the reconcile loop for the CDIConfig object.
 func (r *ImportReconciler) Reconcile(_ context.Context, req reconcile.Request) (reconcile.Result, error) {
 	log := r.log.WithValues("PVC", req.NamespacedName)
@@ -249,11 +287,9 @@ func (r *ImportReconciler) findImporterPod(pvc *corev1.PersistentVolumeClaim, lo
 		}
 		return nil, nil
 	}
-
-	if !metav1.IsControlledBy(pod, pvc) {
+	if !metav1.IsControlledBy(pod, pvc) && !isImageStream(pvc) {
 		return nil, errors.Errorf("Pod is not owned by PVC")
 	}
-
 	log.V(1).Info("Pod is owned by PVC", pod.Name, pvc.Name)
 	return pod, nil
 }
@@ -301,7 +337,7 @@ func (r *ImportReconciler) reconcilePvc(pvc *corev1.PersistentVolumeClaim, log l
 	} else {
 		if pvc.DeletionTimestamp != nil {
 			log.V(1).Info("PVC being terminated, delete pods", "pod.Name", pod.Name)
-			if err := r.client.Delete(context.TODO(), pod); IgnoreNotFound(err) != nil {
+			if err := r.cleanup(pvc, pod, log); err != nil {
 				return reconcile.Result{}, err
 			}
 		} else {
@@ -411,7 +447,7 @@ func (r *ImportReconciler) updatePvcFromPod(pvc *corev1.PersistentVolumeClaim, p
 		}
 		if shouldDeletePod(pvc) {
 			log.V(1).Info("Deleting pod", "pod.Name", pod.Name)
-			if err := r.client.Delete(context.TODO(), pod); IgnoreNotFound(err) != nil {
+			if err := r.cleanup(pvc, pod, log); err != nil {
 				return err
 			}
 		}
@@ -419,6 +455,19 @@ func (r *ImportReconciler) updatePvcFromPod(pvc *corev1.PersistentVolumeClaim, p
 	return nil
 }
 
+func (r *ImportReconciler) cleanup(pvc *corev1.PersistentVolumeClaim, pod *corev1.Pod, log logr.Logger) error {
+	if err := r.client.Delete(context.TODO(), pod); IgnoreNotFound(err) != nil {
+		return err
+	}
+	if HasFinalizer(pvc, importPodImageStreamFinalizer) {
+		RemoveFinalizer(pvc, importPodImageStreamFinalizer)
+		if err := r.updatePVC(pvc, log); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
 func (r *ImportReconciler) updatePVC(pvc *corev1.PersistentVolumeClaim, log logr.Logger) error {
 	log.V(1).Info("Annotations are now", "pvc.anno", pvc.GetAnnotations())
 	if err := r.client.Update(context.TODO(), pvc); err != nil {
@@ -441,16 +490,23 @@ func (r *ImportReconciler) createImporterPod(pvc *corev1.PersistentVolumeClaim)
 
 	if getSource(pvc) == SourceVDDK {
 		r.log.V(1).Info("Pod requires VDDK sidecar for VMware transfer")
-		vddkImageName, err = r.getVddkImageName()
-		if err != nil {
-			anno := pvc.GetAnnotations()
-			anno[AnnBoundCondition] = "false"
-			anno[AnnBoundConditionMessage] = fmt.Sprintf("waiting for %s configmap for VDDK image", common.VddkConfigMap)
-			anno[AnnBoundConditionReason] = common.AwaitingVDDK
-			if updateErr := r.updatePVC(pvc, r.log); updateErr != nil {
-				return updateErr
+		anno := pvc.GetAnnotations()
+		if imageName, ok := anno[AnnVddkInitImageURL]; ok {
+			vddkImageName = &imageName
+		} else {
+			if vddkImageName, err = r.getVddkImageName(); err != nil {
+				r.log.V(1).Error(err, "failed to get VDDK image name from configmap")
 			}
-			return err
+		}
+		if vddkImageName == nil {
+			message := fmt.Sprintf("waiting for %s configmap or %s annotation for VDDK image", common.VddkConfigMap, AnnVddkInitImageURL)
+			anno[AnnBoundCondition] = "false"
+			anno[AnnBoundConditionMessage] = message
+			anno[AnnBoundConditionReason] = common.AwaitingVDDK
+			if err := r.updatePVC(pvc, r.log); err != nil {
+				return err
+			}
+			return errors.New(message)
 		}
 	}
 
@@ -458,14 +514,25 @@ func (r *ImportReconciler) createImporterPod(pvc *corev1.PersistentVolumeClaim)
 	if err != nil {
 		return err
 	}
+	// all checks passed, let's create the importer pod!
+	podArgs := &importerPodArgs{
+		image:             r.image,
+		verbose:           r.verbose,
+		pullPolicy:        r.pullPolicy,
+		podEnvVar:         podEnvVar,
+		pvc:               pvc,
+		scratchPvcName:    scratchPvcName,
+		vddkImageName:     vddkImageName,
+		priorityClassName: getPriorityClass(pvc),
+	}
 
 	// Check for old terminating scratch spaces, this can happen in warm migration scenarios where the scratch
 	// space for a previous pod has not been deleted yet. In that case check if the scratch PVC is terminating
 	// before creating the new importer pod. This should give the PVC time to get deleted.
-	if requiresScratch && scratchPvcName != nil {
+	if requiresScratch && podArgs.scratchPvcName != nil {
 		r.log.V(1).Info("Checking for old scratch pvcs")
 		scratchPvc := &corev1.PersistentVolumeClaim{}
-		err := r.client.Get(context.TODO(), types.NamespacedName{Namespace: pvc.GetNamespace(), Name: *scratchPvcName}, scratchPvc)
+		err := r.client.Get(context.TODO(), types.NamespacedName{Namespace: pvc.GetNamespace(), Name: *podArgs.scratchPvcName}, scratchPvc)
 		if IgnoreNotFound(err) != nil {
 			return err
 		}
@@ -475,12 +542,21 @@ func (r *ImportReconciler) createImporterPod(pvc *corev1.PersistentVolumeClaim)
 		}
 	}
 
-	// all checks passed, let's create the importer pod!
-	pod, err := createImporterPod(r.log, r.client, r.image, r.verbose, r.pullPolicy, podEnvVar, pvc, scratchPvcName, vddkImageName, getPriorityClass(pvc), r.installerLabels)
+	pod, err := createImporterPod(r.log, r.client, podArgs, r.installerLabels)
 	if err != nil {
 		return err
 	}
 	r.log.V(1).Info("Created POD", "pod.Name", pod.Name)
+
+	// If importing from image stream, add finalizer. Note we don't watch the importer pod in this case,
+	// so to prevent a deadlock we add finalizer only if the pod is not retained after completion.
+	if isImageStream(pvc) && pvc.GetAnnotations()[AnnPodRetainAfterCompletion] != "true" {
+		AddFinalizer(pvc, importPodImageStreamFinalizer)
+		if err := r.updatePVC(pvc, r.log); err != nil {
+			return err
+		}
+	}
+
 	if requiresScratch {
 		r.log.V(1).Info("Pod requires scratch space")
 		return r.createScratchPvcForPod(pvc, pod)
@@ -523,6 +599,15 @@ func (r *ImportReconciler) createImportEnvVar(pvc *corev1.PersistentVolumeClaim)
 		podEnvVar.currentCheckpoint = getValueFromAnnotation(pvc, AnnCurrentCheckpoint)
 		podEnvVar.finalCheckpoint = getValueFromAnnotation(pvc, AnnFinalCheckpoint)
 
+		for annotation, value := range pvc.Annotations {
+			if strings.HasPrefix(annotation, AnnExtraHeaders) {
+				podEnvVar.extraHeaders = append(podEnvVar.extraHeaders, value)
+			}
+			if strings.HasPrefix(annotation, AnnSecretExtraHeaders) {
+				podEnvVar.secretExtraHeaders = append(podEnvVar.secretExtraHeaders, value)
+			}
+		}
+
 		var field string
 		if field, err = GetImportProxyConfig(cdiConfig, common.ImportProxyHTTP); err != nil {
 			r.log.V(3).Info("no proxy http url will be supplied:", err.Error())
@@ -561,12 +646,16 @@ func (r *ImportReconciler) createImportEnvVar(pvc *corev1.PersistentVolumeClaim)
 }
 
 func (r *ImportReconciler) isInsecureTLS(pvc *corev1.PersistentVolumeClaim, cdiConfig *cdiv1.CDIConfig) (bool, error) {
-	value, ok := pvc.Annotations[AnnEndpoint]
-	if !ok || value == "" {
+	ep, ok := pvc.Annotations[AnnEndpoint]
+	if !ok || ep == "" {
 		return false, nil
 	}
+	return IsInsecureTLS(ep, cdiConfig, r.uncachedClient, r.log)
+}
 
-	url, err := url.Parse(value)
+// IsInsecureTLS checks if TLS security is disabled for the given endpoint
+func IsInsecureTLS(ep string, cdiConfig *cdiv1.CDIConfig, client client.Client, log logr.Logger) (bool, error) {
+	url, err := url.Parse(ep)
 	if err != nil {
 		return false, err
 	}
@@ -576,7 +665,7 @@ func (r *ImportReconciler) isInsecureTLS(pvc *corev1.PersistentVolumeClaim, cdiC
 	}
 
 	for _, value := range cdiConfig.Spec.InsecureRegistries {
-		r.log.V(1).Info("Checking host against value", "host", url.Host, "value", value)
+		log.V(1).Info("Checking host against value", "host", url.Host, "value", value)
 		if value == url.Host {
 			return true, nil
 		}
@@ -584,19 +673,19 @@ func (r *ImportReconciler) isInsecureTLS(pvc *corev1.PersistentVolumeClaim, cdiC
 
 	// ConfigMap is obsoleted and supported only for upgrade. It won't be refered anymore by future releases.
 	configMapName := common.InsecureRegistryConfigMap
-	r.log.V(1).Info("Checking configmap for host", "configMapName", configMapName, "host URL", url.Host)
+	log.V(1).Info("Checking configmap for host", "configMapName", configMapName, "host URL", url.Host)
 
 	cm := &corev1.ConfigMap{}
-	if err := r.uncachedClient.Get(context.TODO(), types.NamespacedName{Name: configMapName, Namespace: util.GetNamespace()}, cm); err != nil {
+	if err := client.Get(context.TODO(), types.NamespacedName{Name: configMapName, Namespace: util.GetNamespace()}, cm); err != nil {
 		if k8serrors.IsNotFound(err) {
-			r.log.V(1).Info("Configmap does not exist", "configMapName", configMapName)
+			log.V(1).Info("Configmap does not exist", "configMapName", configMapName)
 			return false, nil
 		}
 		return false, err
 	}
 
 	for _, value := range cm.Data {
-		r.log.V(1).Info("Checking host against value", "host", url.Host, "value", value)
+		log.V(1).Info("Checking host against value", "host", url.Host, "value", value)
 		if value == url.Host {
 			return true, nil
 		}
@@ -658,7 +747,7 @@ func (r *ImportReconciler) requiresScratchSpace(pvc *corev1.PersistentVolumeClai
 				scratchRequired = val != ""
 			}
 		case SourceRegistry:
-			scratchRequired = true
+			scratchRequired = pvc.Annotations[AnnRegistryImportMethod] != string(cdiv1.RegistryPullNode)
 		}
 	}
 	value, ok := pvc.Annotations[AnnRequiresScratch]
@@ -685,7 +774,7 @@ func (r *ImportReconciler) createScratchPvcForPod(pvc *corev1.PersistentVolumeCl
 
 		storageClassName := GetScratchPvcStorageClass(r.client, pvc)
 		// Scratch PVC doesn't exist yet, create it. Determine which storage class to use.
-		_, err = CreateScratchPersistentVolumeClaim(r.client, pvc, pod, scratchPVCName, storageClassName, r.installerLabels)
+		_, err = CreateScratchPersistentVolumeClaim(r.client, pvc, pod, scratchPVCName, storageClassName, r.installerLabels, r.recorder)
 		if err != nil {
 			return err
 		}
@@ -769,12 +858,37 @@ func getEndpoint(pvc *corev1.PersistentVolumeClaim) (string, error) {
 	return ep, nil
 }
 
+// returns the import image part of the endpoint string
+func getRegistryImportImage(pvc *corev1.PersistentVolumeClaim) (string, error) {
+	ep, err := getEndpoint(pvc)
+	if err != nil {
+		return "", nil
+	}
+	if isImageStream(pvc) {
+		return ep, nil
+	}
+	url, err := url.Parse(ep)
+	if err != nil {
+		return "", errors.Errorf("illegal registry endpoint %s", ep)
+	}
+	return url.Host + url.Path, nil
+}
+
 // getValueFromAnnotation returns the value of an annotation
 func getValueFromAnnotation(pvc *corev1.PersistentVolumeClaim, annotation string) string {
 	value, _ := pvc.Annotations[annotation]
 	return value
 }
 
+// If this pod is going to transfer one checkpoint in a multi-stage import, attach the checkpoint name to the pod name so
+// that each checkpoint gets a unique pod. That way each pod can be inspected using the retainAfterCompletion annotation.
+func podNameWithCheckpoint(pvc *corev1.PersistentVolumeClaim) string {
+	if checkpoint := pvc.Annotations[AnnCurrentCheckpoint]; checkpoint != "" {
+		return pvc.Name + "-checkpoint-" + checkpoint
+	}
+	return pvc.Name
+}
+
 func getImportPodNameFromPvc(pvc *corev1.PersistentVolumeClaim) string {
 	podName, ok := pvc.Annotations[AnnImportPod]
 	if ok {
@@ -782,41 +896,172 @@ func getImportPodNameFromPvc(pvc *corev1.PersistentVolumeClaim) string {
 	}
 	// fallback to legacy naming, in fact the following function is fully compatible with legacy
 	// name concatenation "importer-{pvc.Name}" if the name length is under the size limits,
-	return naming.GetResourceName(common.ImporterPodName, pvc.Name)
+	return naming.GetResourceName(common.ImporterPodName, podNameWithCheckpoint(pvc))
 }
 
 func createImportPodNameFromPvc(pvc *corev1.PersistentVolumeClaim) string {
-	return naming.GetResourceName(common.ImporterPodName, pvc.Name)
+	return naming.GetResourceName(common.ImporterPodName, podNameWithCheckpoint(pvc))
 }
 
 // createImporterPod creates and returns a pointer to a pod which is created based on the passed-in endpoint, secret
 // name, and pvc. A nil secret means the endpoint credentials are not passed to the
 // importer pod.
-func createImporterPod(log logr.Logger, client client.Client, image, verbose, pullPolicy string, podEnvVar *importPodEnvVar, pvc *corev1.PersistentVolumeClaim, scratchPvcName *string, vddkImageName *string, priorityClassName string, installerLabels map[string]string) (*corev1.Pod, error) {
-	podResourceRequirements, err := GetDefaultPodResourceRequirements(client)
+func createImporterPod(log logr.Logger, client client.Client, args *importerPodArgs, installerLabels map[string]string) (*corev1.Pod, error) {
+	var err error
+	args.podResourceRequirements, err = GetDefaultPodResourceRequirements(client)
 	if err != nil {
 		return nil, err
 	}
 
-	workloadNodePlacement, err := GetWorkloadNodePlacement(client)
+	args.workloadNodePlacement, err = GetWorkloadNodePlacement(client)
 	if err != nil {
 		return nil, err
 	}
 
-	pod := makeImporterPodSpec(pvc.Namespace, image, verbose, pullPolicy, podEnvVar, pvc, scratchPvcName, podResourceRequirements, workloadNodePlacement, vddkImageName, priorityClassName)
+	var pod *corev1.Pod
+	if getSource(args.pvc) == SourceRegistry && args.pvc.Annotations[AnnRegistryImportMethod] == string(cdiv1.RegistryPullNode) {
+		args.importImage, err = getRegistryImportImage(args.pvc)
+		if err != nil {
+			return nil, err
+		}
+		pod = makeNodeImporterPodSpec(args)
+	} else {
+		pod = makeImporterPodSpec(args)
+	}
+
 	util.SetRecommendedLabels(pod, installerLabels, "cdi-controller")
 
-	if err := client.Create(context.TODO(), pod); err != nil {
+	if err = client.Create(context.TODO(), pod); err != nil {
 		return nil, err
 	}
-	log.V(3).Info("importer pod created\n", "pod.Name", pod.Name, "pod.Namespace", pod.Namespace, "image name", image)
+
+	log.V(3).Info("importer pod created\n", "pod.Name", pod.Name, "pod.Namespace", pod.Namespace, "image name", args.image)
 	return pod, nil
 }
 
-// makeImporterPodSpec creates and return the importer pod spec based on the passed-in endpoint, secret and pvc.
-func makeImporterPodSpec(namespace, image, verbose, pullPolicy string, podEnvVar *importPodEnvVar, pvc *corev1.PersistentVolumeClaim, scratchPvcName *string, podResourceRequirements *corev1.ResourceRequirements, workloadNodePlacement *sdkapi.NodePlacement, vddkImageName *string, priorityClassName string) *corev1.Pod {
+// makeNodeImporterPodSpec creates and returns the node docker cache based importer pod spec based on the passed-in importImage and pvc.
+func makeNodeImporterPodSpec(args *importerPodArgs) *corev1.Pod {
 	// importer pod name contains the pvc name
-	podName, _ := pvc.Annotations[AnnImportPod]
+	podName, _ := args.pvc.Annotations[AnnImportPod]
+
+	volumes := []corev1.Volume{
+		{
+			Name: "shared-volume",
+			VolumeSource: corev1.VolumeSource{
+				EmptyDir: &corev1.EmptyDirVolumeSource{},
+			},
+		},
+		{
+			Name: DataVolName,
+			VolumeSource: corev1.VolumeSource{
+				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
+					ClaimName: args.pvc.Name,
+					ReadOnly:  false,
+				},
+			},
+		},
+	}
+
+	importerContainer := makeImporterContainerSpec(args.image, args.verbose, args.pullPolicy)
+
+	pod := &corev1.Pod{
+		TypeMeta: metav1.TypeMeta{
+			Kind:       "Pod",
+			APIVersion: "v1",
+		},
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      podName,
+			Namespace: args.pvc.Namespace,
+			Annotations: map[string]string{
+				AnnCreatedBy: "yes",
+			},
+			Labels: map[string]string{
+				common.CDILabelKey:        common.CDILabelValue,
+				common.CDIComponentLabel:  common.ImporterPodName,
+				common.PrometheusLabelKey: common.PrometheusLabelValue,
+			},
+		},
+		Spec: corev1.PodSpec{
+			InitContainers: []corev1.Container{
+				{
+					Name:            "init",
+					Image:           args.image,
+					ImagePullPolicy: corev1.PullPolicy(args.pullPolicy),
+					Command:         []string{"sh", "-c", "cp /usr/bin/cdi-containerimage-server /shared/server"},
+					VolumeMounts: []corev1.VolumeMount{
+						{
+							MountPath: "/shared",
+							Name:      "shared-volume",
+						},
+					},
+				},
+			},
+			Containers: []corev1.Container{
+				*importerContainer,
+				{
+					Name:            "server",
+					Image:           args.importImage,
+					ImagePullPolicy: corev1.PullPolicy(args.pullPolicy),
+					Command:         []string{"/shared/server", "-p", "8100", "-image-dir", "/disk", "-ready-file", "/shared/ready", "-done-file", "/shared/done"},
+					VolumeMounts: []corev1.VolumeMount{
+						{
+							MountPath: "/shared",
+							Name:      "shared-volume",
+						},
+					},
+				},
+			},
+			RestartPolicy:     corev1.RestartPolicyOnFailure,
+			Volumes:           volumes,
+			NodeSelector:      args.workloadNodePlacement.NodeSelector,
+			Tolerations:       args.workloadNodePlacement.Tolerations,
+			Affinity:          args.workloadNodePlacement.Affinity,
+			PriorityClassName: args.priorityClassName,
+		},
+	}
+
+	/**
+	FIXME: When registry source is ImageStream, if we set importer pod OwnerReference (to its pvc, like all other cases),
+	for some reason (OCP issue?) we get the following error:
+		Failed to pull image "imagestream-name": rpc error: code = Unknown
+		desc = Error reading manifest latest in docker.io/library/imagestream-name: errors:
+		denied: requested access to the resource is denied
+		unauthorized: authentication required
+	When we don't set pod OwnerReferences, all works well.
+	*/
+	if isImageStream(args.pvc) {
+		pod.Annotations[AnnOpenShiftImageLookup] = "*"
+	} else {
+		blockOwnerDeletion := true
+		isController := true
+		ownerRef := metav1.OwnerReference{
+			APIVersion:         "v1",
+			Kind:               "PersistentVolumeClaim",
+			Name:               args.pvc.Name,
+			UID:                args.pvc.GetUID(),
+			BlockOwnerDeletion: &blockOwnerDeletion,
+			Controller:         &isController,
+		}
+		pod.OwnerReferences = append(pod.OwnerReferences, ownerRef)
+	}
+
+	args.podEnvVar.source = SourceHTTP
+	args.podEnvVar.ep = "http://localhost:8100/disk.img"
+	args.podEnvVar.readyFile = "/shared/ready"
+	args.podEnvVar.doneFile = "/shared/done"
+	setImporterPodCommons(pod, args.podEnvVar, args.pvc, args.podResourceRequirements)
+	pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, corev1.VolumeMount{
+		MountPath: "/shared",
+		Name:      "shared-volume",
+	})
+
+	return pod
+}
+
+// makeImporterPodSpec creates and return the importer pod spec based on the passed-in endpoint, secret and pvc.
+func makeImporterPodSpec(args *importerPodArgs) *corev1.Pod {
+	// importer pod name contains the pvc name
+	podName, _ := args.pvc.Annotations[AnnImportPod]
 
 	blockOwnerDeletion := true
 	isController := true
@@ -826,25 +1071,27 @@ func makeImporterPodSpec(namespace, image, verbose, pullPolicy string, podEnvVar
 			Name: DataVolName,
 			VolumeSource: corev1.VolumeSource{
 				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
-					ClaimName: pvc.Name,
+					ClaimName: args.pvc.Name,
 					ReadOnly:  false,
 				},
 			},
 		},
 	}
 
-	if scratchPvcName != nil {
+	if args.scratchPvcName != nil {
 		volumes = append(volumes, corev1.Volume{
 			Name: ScratchVolName,
 			VolumeSource: corev1.VolumeSource{
 				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
-					ClaimName: *scratchPvcName,
+					ClaimName: *args.scratchPvcName,
 					ReadOnly:  false,
 				},
 			},
 		})
 	}
 
+	importerContainer := makeImporterContainerSpec(args.image, args.verbose, args.pullPolicy)
+
 	pod := &corev1.Pod{
 		TypeMeta: metav1.TypeMeta{
 			Kind:       "Pod",
@@ -852,21 +1099,21 @@ func makeImporterPodSpec(namespace, image, verbose, pullPolicy string, podEnvVar
 		},
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      podName,
-			Namespace: namespace,
+			Namespace: args.pvc.Namespace,
 			Annotations: map[string]string{
 				AnnCreatedBy: "yes",
 			},
 			Labels: map[string]string{
-				common.CDILabelKey:       common.CDILabelValue,
-				common.CDIComponentLabel: common.ImporterPodName,
-				common.PrometheusLabel:   "",
+				common.CDILabelKey:        common.CDILabelValue,
+				common.CDIComponentLabel:  common.ImporterPodName,
+				common.PrometheusLabelKey: common.PrometheusLabelValue,
 			},
 			OwnerReferences: []metav1.OwnerReference{
 				{
 					APIVersion:         "v1",
 					Kind:               "PersistentVolumeClaim",
-					Name:               pvc.Name,
-					UID:                pvc.GetUID(),
+					Name:               args.pvc.Name,
+					UID:                args.pvc.GetUID(),
 					BlockOwnerDeletion: &blockOwnerDeletion,
 					Controller:         &isController,
 				},
@@ -874,31 +1121,106 @@ func makeImporterPodSpec(namespace, image, verbose, pullPolicy string, podEnvVar
 		},
 		Spec: corev1.PodSpec{
 			Containers: []corev1.Container{
-				{
-					Name:            common.ImporterPodName,
-					Image:           image,
-					ImagePullPolicy: corev1.PullPolicy(pullPolicy),
-					Args:            []string{"-v=" + verbose},
-					Ports: []corev1.ContainerPort{
-						{
-							Name:          "metrics",
-							ContainerPort: 8443,
-							Protocol:      corev1.ProtocolTCP,
-						},
-					},
-				},
+				*importerContainer,
 			},
 			RestartPolicy:     corev1.RestartPolicyOnFailure,
 			Volumes:           volumes,
-			NodeSelector:      workloadNodePlacement.NodeSelector,
-			Tolerations:       workloadNodePlacement.Tolerations,
-			Affinity:          workloadNodePlacement.Affinity,
-			PriorityClassName: priorityClassName,
+			NodeSelector:      args.workloadNodePlacement.NodeSelector,
+			Tolerations:       args.workloadNodePlacement.Tolerations,
+			Affinity:          args.workloadNodePlacement.Affinity,
+			PriorityClassName: args.priorityClassName,
 		},
 	}
 
+	setImporterPodCommons(pod, args.podEnvVar, args.pvc, args.podResourceRequirements)
+
+	if args.scratchPvcName != nil {
+		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, corev1.VolumeMount{
+			Name:      ScratchVolName,
+			MountPath: common.ScratchDataDir,
+		})
+	}
+
+	if args.vddkImageName != nil {
+		pod.Spec.Volumes = append(pod.Spec.Volumes, corev1.Volume{
+			Name: "vddk-vol-mount",
+			VolumeSource: corev1.VolumeSource{
+				EmptyDir: &corev1.EmptyDirVolumeSource{},
+			},
+		})
+		pod.Spec.InitContainers = append(pod.Spec.InitContainers, corev1.Container{
+			Name:  "vddk-side-car",
+			Image: *args.vddkImageName,
+			VolumeMounts: []corev1.VolumeMount{
+				{
+					Name:      "vddk-vol-mount",
+					MountPath: "/opt",
+				},
+			},
+		})
+		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, corev1.VolumeMount{
+			Name:      "vddk-vol-mount",
+			MountPath: "/opt",
+		})
+	}
+
+	if args.podEnvVar.certConfigMap != "" {
+		vm := corev1.VolumeMount{
+			Name:      CertVolName,
+			MountPath: common.ImporterCertDir,
+		}
+
+		vol := corev1.Volume{
+			Name: CertVolName,
+			VolumeSource: corev1.VolumeSource{
+				ConfigMap: &corev1.ConfigMapVolumeSource{
+					LocalObjectReference: corev1.LocalObjectReference{
+						Name: args.podEnvVar.certConfigMap,
+					},
+				},
+			},
+		}
+
+		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, vm)
+		pod.Spec.Volumes = append(pod.Spec.Volumes, vol)
+	}
+
+	if args.podEnvVar.certConfigMapProxy != "" {
+		vm := corev1.VolumeMount{
+			Name:      ProxyCertVolName,
+			MountPath: common.ImporterProxyCertDir,
+		}
+		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, vm)
+		pod.Spec.Volumes = append(pod.Spec.Volumes, createProxyConfigMapVolume(ProxyCertVolName, args.podEnvVar.certConfigMapProxy))
+	}
+
+	for index, header := range args.podEnvVar.secretExtraHeaders {
+		vm := corev1.VolumeMount{
+			Name:      fmt.Sprintf(SecretExtraHeadersVolumeName, index),
+			MountPath: path.Join(common.ImporterSecretExtraHeadersDir, fmt.Sprint(index)),
+		}
+
+		vol := corev1.Volume{
+			Name: fmt.Sprintf(SecretExtraHeadersVolumeName, index),
+			VolumeSource: corev1.VolumeSource{
+				Secret: &corev1.SecretVolumeSource{
+					SecretName: header,
+				},
+			},
+		}
+
+		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, vm)
+		pod.Spec.Volumes = append(pod.Spec.Volumes, vol)
+	}
+
+	return pod
+}
+
+func setImporterPodCommons(pod *corev1.Pod, podEnvVar *importPodEnvVar, pvc *corev1.PersistentVolumeClaim, podResourceRequirements *corev1.ResourceRequirements) {
 	if podResourceRequirements != nil {
-		pod.Spec.Containers[0].Resources = *podResourceRequirements
+		for i := range pod.Spec.Containers {
+			pod.Spec.Containers[i].Resources = *podResourceRequirements
+		}
 	}
 
 	ownerUID := pvc.UID
@@ -915,68 +1237,8 @@ func makeImporterPodSpec(namespace, image, verbose, pullPolicy string, podEnvVar
 		pod.Spec.Containers[0].VolumeMounts = addImportVolumeMounts()
 	}
 
-	if scratchPvcName != nil {
-		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, corev1.VolumeMount{
-			Name:      ScratchVolName,
-			MountPath: common.ScratchDataDir,
-		})
-	}
-
-	if vddkImageName != nil {
-		pod.Spec.Volumes = append(pod.Spec.Volumes, corev1.Volume{
-			Name: "vddk-vol-mount",
-			VolumeSource: corev1.VolumeSource{
-				EmptyDir: &corev1.EmptyDirVolumeSource{},
-			},
-		})
-		pod.Spec.InitContainers = append(pod.Spec.InitContainers, corev1.Container{
-			Name:  "vddk-side-car",
-			Image: *vddkImageName,
-			VolumeMounts: []corev1.VolumeMount{
-				{
-					Name:      "vddk-vol-mount",
-					MountPath: "/opt",
-				},
-			},
-		})
-		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, corev1.VolumeMount{
-			Name:      "vddk-vol-mount",
-			MountPath: "/opt",
-		})
-	}
-
 	pod.Spec.Containers[0].Env = makeImportEnv(podEnvVar, ownerUID)
 
-	if podEnvVar.certConfigMap != "" {
-		vm := corev1.VolumeMount{
-			Name:      CertVolName,
-			MountPath: common.ImporterCertDir,
-		}
-
-		vol := corev1.Volume{
-			Name: CertVolName,
-			VolumeSource: corev1.VolumeSource{
-				ConfigMap: &corev1.ConfigMapVolumeSource{
-					LocalObjectReference: corev1.LocalObjectReference{
-						Name: podEnvVar.certConfigMap,
-					},
-				},
-			},
-		}
-
-		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, vm)
-		pod.Spec.Volumes = append(pod.Spec.Volumes, vol)
-	}
-
-	if podEnvVar.certConfigMapProxy != "" {
-		vm := corev1.VolumeMount{
-			Name:      ProxyCertVolName,
-			MountPath: common.ImporterProxyCertDir,
-		}
-		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, vm)
-		pod.Spec.Volumes = append(pod.Spec.Volumes, createProxyConfigMapVolume(ProxyCertVolName, podEnvVar.certConfigMapProxy))
-	}
-
 	if podEnvVar.contentType == string(cdiv1.DataVolumeKubeVirt) {
 		// Set the fsGroup on the security context to the QemuSubGid
 		if pod.Spec.SecurityContext == nil {
@@ -986,7 +1248,22 @@ func makeImporterPodSpec(namespace, image, verbose, pullPolicy string, podEnvVar
 		pod.Spec.SecurityContext.FSGroup = &fsGroup
 	}
 	SetPodPvcAnnotations(pod, pvc)
-	return pod
+}
+
+func makeImporterContainerSpec(image, verbose, pullPolicy string) *corev1.Container {
+	return &corev1.Container{
+		Name:            common.ImporterPodName,
+		Image:           image,
+		ImagePullPolicy: corev1.PullPolicy(pullPolicy),
+		Args:            []string{"-v=" + verbose},
+		Ports: []corev1.ContainerPort{
+			{
+				Name:          "metrics",
+				ContainerPort: 8443,
+				Protocol:      corev1.ProtocolTCP,
+			},
+		},
+	}
 }
 
 func createProxyConfigMapVolume(certVolName, objRef string) corev1.Volume {
@@ -1052,6 +1329,14 @@ func makeImportEnv(podEnvVar *importPodEnvVar, uid types.UID) []corev1.EnvVar {
 			Name:  common.ImporterUUID,
 			Value: podEnvVar.uuid,
 		},
+		{
+			Name:  common.ImporterReadyFile,
+			Value: podEnvVar.readyFile,
+		},
+		{
+			Name:  common.ImporterDoneFile,
+			Value: podEnvVar.doneFile,
+		},
 		{
 			Name:  common.ImporterBackingFile,
 			Value: podEnvVar.backingFile,
@@ -1125,5 +1410,11 @@ func makeImportEnv(podEnvVar *importPodEnvVar, uid types.UID) []corev1.EnvVar {
 			Value: common.ImporterProxyCertDir,
 		})
 	}
+	for index, header := range podEnvVar.extraHeaders {
+		env = append(env, corev1.EnvVar{
+			Name:  fmt.Sprintf("%s%d", common.ImporterExtraHeader, index),
+			Value: header,
+		})
+	}
 	return env
 }
