@@ -18,8 +18,9 @@ package cluster
 
 import (
 	rbacv1 "k8s.io/api/rbac/v1"
-	"kubevirt.io/containerized-data-importer/pkg/operator/resources/utils"
 	"sigs.k8s.io/controller-runtime/pkg/client"
+
+	"kubevirt.io/containerized-data-importer/pkg/operator/resources/utils"
 )
 
 const (
@@ -129,6 +130,7 @@ func getControllerClusterPolicyRules() []rbacv1.PolicyRule {
 			},
 			Resources: []string{
 				"storageclasses",
+				"csidrivers",
 			},
 			Verbs: []string{
 				"get",
@@ -210,6 +212,54 @@ func getControllerClusterPolicyRules() []rbacv1.PolicyRule {
 				"watch",
 			},
 		},
+		{
+			APIGroups: []string{
+				"image.openshift.io",
+			},
+			Resources: []string{
+				"imagestreams",
+			},
+			Verbs: []string{
+				"get",
+				"list",
+				"watch",
+			},
+		},
+		{
+			APIGroups: []string{
+				"",
+			},
+			Resources: []string{
+				"secrets",
+			},
+			Verbs: []string{
+				"create",
+			},
+		},
+		{
+			APIGroups: []string{
+				"batch",
+			},
+			Resources: []string{
+				"cronjobs",
+			},
+			Verbs: []string{
+				"list",
+				"watch",
+			},
+		},
+		{
+			APIGroups: []string{
+				"batch",
+			},
+			Resources: []string{
+				"jobs",
+			},
+			Verbs: []string{
+				"list",
+				"watch",
+			},
+		},
 	}
 }
 
